@@ -15,9 +15,16 @@ class AbstractOffPolicyLoss(Generic[ModelType], Loss):
     """Abstract class representing an off-policy loss function used to train a
     model."""
 
+    # noinspection PyMethodOverriding
     @abc.abstractmethod
     def loss(  # type: ignore
-        self, model: ModelType, batch: ObservationType, memory: Memory, *args, **kwargs,
+        self,
+        *,  # No positional arguments
+        step_count: int,
+        model: ModelType,
+        batch: ObservationType,
+        memory: Memory,
+        **kwargs,
     ) -> Tuple[torch.FloatTensor, Dict[str, float], Memory, int]:
         """Computes the loss.
 
