@@ -2,6 +2,8 @@
 # Modified work Copyright (c) Allen Institute for AI
 # This source code is licensed under the MIT license found in the
 # LICENSE file in the root directory of this source tree.
+import os
+import signal
 import time
 import traceback
 from multiprocessing.connection import Connection
@@ -33,7 +35,6 @@ from allenact.utils.misc_utils import partition_sequence
 from allenact.utils.system import get_logger
 from allenact.utils.tensor_utils import tile_images
 
-
 try:
     # Use torch.multiprocessing if we can.
     # We have yet to find a reason to not use it and
@@ -62,7 +63,55 @@ PAUSE_COMMAND = "pause"
 RESUME_COMMAND = "resume"
 
 
-class VectorSampledTasks(object):
+class DelaySignalHandling:
+    # Modified from https://stackoverflow.com/a/21919644
+    def __init__(self):
+        self.int_signal_received: Optional[Any] = None
+        self.term_signal_received: Optional[Any] = None
+        self.old_int_handler = None
+        self.old_term_handler = None
+
+    def __enter__(self):
+        self.int_signal_received: Optional[Any] = None
+        self.term_signal_received: Optional[Any] = None
+        self.old_int_handler = signal.signal(signal.SIGINT, self.int_handler)
+        self.old_term_handler = signal.signal(signal.SIGTERM, self.term_handler)
+
+    def int_handler(self, sig, frame):
+        self.int_signal_received = (sig, frame)
+        get_logger().debug("SIGINT received. Delaying KeyboardInterrupt.")
+
+    def term_handler(self, sig, frame):
+        self.term_signal_received = (sig, frame)
+        get_logger().debug("SIGTERM received. Delaying termination.")
+
+    def __exit__(self, type, value, traceback):
+        signal.signal(signal.SIGINT, self.old_int_handler)
+        signal.signal(signal.SIGTERM, self.old_term_handler)
+        if self.term_signal_received:
+            # For some reason there appear to be cases where the original termination
+            # handler is not callable. It is unclear to me exactly why this is the case
+            # but here we add a guard to double check that the handler is callable and,
+            # if it's not, we re-send the termination signal to the process and let
+            # the python internals handle it (note that we've already reset the termination
+            # handler to what it was originaly above in the signal.signal(...) code).
+            if callable(self.old_term_handler):
+                self.old_term_handler(*self.term_signal_received)
+            else:
+                get_logger().debug(
+                    "Termination handler could not be called after delaying signal handling."
+                    f" Resending the SIGTERM signal. Last (sig, frame) == ({self.term_signal_received})."
+                )
+                os.kill(os.getpid(), signal.SIGTERM)
+
+        if self.int_signal_received:
+            if callable(self.old_int_handler):
+                self.old_int_handler(*self.int_signal_received)
+            else:
+                signal.default_int_handler(*self.int_signal_received)
+
+
+class VectorSampledTasks:
     """Vectorized collection of tasks. Creates multiple processes where each
     process runs its own TaskSampler. Each process generates one Task from its
     TaskSampler at a time and this class allows for interacting with these
@@ -263,52 +312,61 @@ class VectorSampledTasks(object):
             while True:
                 read_input = connection_read_fn()
 
-                if len(read_input) == 3:
-                    sampler_index, command, data = read_input
+                with DelaySignalHandling():
+                    # Delaying signal handling here is necessary to ensure that we don't
+                    # (when processing a SIGTERM/SIGINT signal) attempt to send data to
+                    # a generator while it is already processing other data.
+                    if len(read_input) == 3:
+                        sampler_index, command, data = read_input
 
-                    assert command != CLOSE_COMMAND, "Must close all processes at once."
-                    assert (
-                        command != RESUME_COMMAND
-                    ), "Must resume all task samplers at once."
+                        assert (
+                            command != CLOSE_COMMAND
+                        ), "Must close all processes at once."
+                        assert (
+                            command != RESUME_COMMAND
+                        ), "Must resume all task samplers at once."
 
-                    if command == PAUSE_COMMAND:
-                        sp_vector_sampled_tasks.pause_at(sampler_index=sampler_index)
-                        connection_write_fn("done")
-                    else:
-                        connection_write_fn(
-                            sp_vector_sampled_tasks.command_at(
-                                sampler_index=sampler_index, command=command, data=data
+                        if command == PAUSE_COMMAND:
+                            sp_vector_sampled_tasks.pause_at(
+                                sampler_index=sampler_index
                             )
-                        )
-                else:
-                    commands, data_list = read_input
-
-                    assert (
-                        commands != PAUSE_COMMAND
-                    ), "Cannot pause all task samplers at once."
-
-                    if commands == CLOSE_COMMAND:
-                        sp_vector_sampled_tasks.close()
-                        break
-                    elif commands == RESUME_COMMAND:
-                        sp_vector_sampled_tasks.resume_all()
-                        connection_write_fn("done")
-                    else:
-                        if isinstance(commands, str):
-                            commands = [
-                                commands
-                            ] * sp_vector_sampled_tasks.num_unpaused_tasks
-
-                        connection_write_fn(
-                            sp_vector_sampled_tasks.command(
-                                commands=commands, data_list=data_list
+                            connection_write_fn("done")
+                        else:
+                            connection_write_fn(
+                                sp_vector_sampled_tasks.command_at(
+                                    sampler_index=sampler_index,
+                                    command=command,
+                                    data=data,
+                                )
                             )
-                        )
+                    else:
+                        commands, data_list = read_input
 
-        except KeyboardInterrupt as e:
+                        assert (
+                            commands != PAUSE_COMMAND
+                        ), "Cannot pause all task samplers at once."
+
+                        if commands == CLOSE_COMMAND:
+                            sp_vector_sampled_tasks.close()
+                            break
+                        elif commands == RESUME_COMMAND:
+                            sp_vector_sampled_tasks.resume_all()
+                            connection_write_fn("done")
+                        else:
+                            if isinstance(commands, str):
+                                commands = [
+                                    commands
+                                ] * sp_vector_sampled_tasks.num_unpaused_tasks
+
+                            connection_write_fn(
+                                sp_vector_sampled_tasks.command(
+                                    commands=commands, data_list=data_list
+                                )
+                            )
+
+        except KeyboardInterrupt:
             if should_log:
-                get_logger().info("Worker {} KeyboardInterrupt".format(worker_id))
-            raise e
+                get_logger().info(f"Worker {worker_id} KeyboardInterrupt")
         except Exception as e:
             get_logger().error(traceback.format_exc())
             raise e
@@ -316,7 +374,7 @@ class VectorSampledTasks(object):
             if child_pipe is not None:
                 child_pipe.close()
             if should_log:
-                get_logger().info("""Worker {} closing.""".format(worker_id))
+                get_logger().info(f"Worker {worker_id} closing.")
 
     def _spawn_workers(
         self,
@@ -536,19 +594,19 @@ class VectorSampledTasks(object):
             for read_fn in self._connection_read_fns:
                 try:
                     read_fn()
-                except:
+                except Exception:
                     pass
 
         for write_fn in self._connection_write_fns:
             try:
                 write_fn((CLOSE_COMMAND, None))
-            except:
+            except Exception:
                 pass
 
         for process in self._workers:
             try:
                 process.join(timeout=0.1)
-            except:
+            except Exception:
                 pass
 
         self._is_closed = True
@@ -864,6 +922,8 @@ class SingleProcessVectorSampledTasks(object):
                     if current_task.is_done():
                         metrics = current_task.metrics()
                         if metrics is not None and len(metrics) != 0:
+                            if step_result.info is None:
+                                step_result = step_result.clone({"info": {}})
                             step_result.info[COMPLETE_TASK_METRICS_KEY] = metrics
 
                         if auto_resample_when_done:
@@ -1085,8 +1145,11 @@ class SingleProcessVectorSampledTasks(object):
 
         for g in self._vector_task_generators:
             try:
-                g.send((CLOSE_COMMAND, None))
-            except StopIteration:
+                try:
+                    g.send((CLOSE_COMMAND, None))
+                except StopIteration:
+                    pass
+            except KeyboardInterrupt:
                 pass
 
         self._is_closed = True
