@@ -13,12 +13,15 @@ from allenact.utils.misc_utils import md5_hash_str_as_int
 def md5_hash_of_state_dict(state_dict: Dict[str, Any]):
     hashables = []
     for piece in sorted(state_dict.items()):
-        if isinstance(piece[1], (torch.Tensor, nn.Parameter)):
-            hashables.append(
-                int(hashlib.md5(piece[1].data.cpu().numpy().tobytes()).hexdigest(), 16,)
-            )
+        if isinstance(piece[1], (np.ndarray, torch.Tensor, nn.Parameter)):
+            hashables.append(piece[0])
+            if not isinstance(piece[1], np.ndarray):
+                p1 = piece[1].data.cpu().numpy()
+            else:
+                p1 = piece[1]
+            hashables.append(int(hashlib.md5(p1.tobytes()).hexdigest(), 16,))
         else:
-            md5_hash_str_as_int(str(piece))
+            hashables.append(md5_hash_str_as_int(str(piece)))
 
     return md5_hash_str_as_int(str(hashables))
 
@@ -26,6 +29,7 @@ def md5_hash_of_state_dict(state_dict: Dict[str, Any]):
 class Flatten(nn.Module):
     """Flatten input tensor so that it is of shape (FLATTENED_BATCH x -1)."""
 
+    # noinspection PyMethodMayBeStatic
     def forward(self, x):
         """Flatten input tensor.
 
@@ -155,7 +159,6 @@ def compute_cnn_output(
 
     CNN output with dimensions [STEP, SAMPLER, AGENT, CHANNEL, (HEIGHT, WIDTH)].
     """
-    use_agents: bool
     nsteps: int
     nsamplers: int
     nagents: int
@@ -172,7 +175,7 @@ def compute_cnn_output(
         nsteps, nsamplers = cnn_input.shape[:2]
 
     # Make FLAT_BATCH = nsteps * nsamplers (* nagents)
-    cnn_input = cnn_input.view((-1,) + cnn_input.shape[2 + int(use_agents) :])
+    cnn_input = cnn_input.view((-1,) + cnn_input.shape[2 + int(nagents is not None) :])
 
     if permute_order is not None:
         cnn_input = cnn_input.permute(*permute_order)
@@ -217,3 +220,24 @@ def simple_linear_weights_init(m):
         m.weight.data.uniform_(-w_bound, w_bound)
         if m.bias is not None:
             m.bias.data.fill_(0)
+
+
+class FeatureEmbedding(nn.Module):
+    """A wrapper of nn.Embedding but support zero output Used for extracting
+    features for actions/rewards."""
+
+    def __init__(self, input_size, output_size):
+        super().__init__()
+        self.input_size = input_size
+        self.output_size = output_size
+        if self.output_size != 0:
+            self.fc = nn.Embedding(input_size, output_size)
+        else:  # automatically be moved to a device
+            self.null_embedding: torch.Tensor
+            self.register_buffer("null_embedding", torch.zeros(0,), persistent=False)
+
+    def forward(self, inputs):
+        if self.output_size != 0:
+            return self.fc(inputs)
+        else:
+            return self.null_embedding
